#!/usr/bin/python
# Copyright (C) 2006 by John Szakmeister
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import os, sys, string

class FsfsError(Exception):
  pass

class FsfsInvalidHeaderError(FsfsError):
  def __init__ (self, msg, noderev, rep, offset):
    self.noderev = noderev
    self.rep = rep
    self.offset = offset
    FsfsError.__init__(self, msg)

class CmdlineError(Exception):
  pass


def create_file_obj(fileobj):
  new_d = os.dup(fileobj.fileno())
  new_obj = os.fdopen(new_d)
  return new_obj


def convert_to_binary_string(s):
  str = ''
  for byte in s:
    cbyte = ord(byte)
    for i in xrange(8):
      if cbyte & (1 << (7 - i)):
        str = str + '1'
      else:
        str = str + '0'
  return str


class bitstream(object):
  def __init__(self, f):
    self.f = f
    self.bits = ''
    self.length = 0

  def _ensure_bits(self, num_bits):
    if len(self.bits) >= num_bits:
      return

    num_bytes = (num_bits - len(self.bits) + 7) / 8;
    self.bits = self.bits + convert_to_binary_string(self.f.read(num_bytes))
    
  def read_bits(self, num_bits):
    self._ensure_bits(num_bits)
    bits = self.bits[:num_bits]
    self.bits = self.bits[num_bits:]
    try:
      return int(bits, 2)
    except:
      raise EOFError, \
        "Unexpected end of file while reading bits (%d, %d, '%s')" % \
        (self.f.tell(), num_bits, bits)


def get_varint(bs):
  i = 0
  while True:
    #byte = bs.read_bits(8)
    byte = ord(bs.f.read(1))
    i = (i << 7) + (byte & 0x7F)
    if byte & 0x80 == 0:
      break
  return i


def parse_svndiff_instruction(bs, dump_instruction):
  start = bs.f.tell()
  instr = bs.read_bits(2)
  length = bs.read_bits(6)
  total_bytes = 0
  data_length = 0
  target_length = 0

  # Make sure we have a valid instruction
  if instr == 0x3:
    raise "Invalid instruction found at offset %d" % start

  if length == 0:
    length = get_varint(bs)

  if instr == 0 or instr == 1:
    offset = get_varint(bs)
  else:
    ### This should be the offset into the new data after
    ### having done the last copy... which obviously means
    ### we need a state machine. :-(
    offset = 0

  if instr == 0:
    instr = 'copy-source'
  elif instr == 1:
    instr = 'copy-target'
  elif instr == 2:
    instr = 'copy-data'

  if dump_instruction:
    print '    instr (%d, %d): %s %d %d' % (start, bs.f.tell() - start,
                                            instr, length, offset)

  # return the length of this instruction and the number of data bytes
  # processed.
  if instr == 'copy-data':
    data_length = length
  else:
    target_length = length
  return (bs.f.tell() - start, data_length, target_length)


def parse_svndiff(f, length, dump_instructions=False, dump_windows=False):
  ### If I ever get to the point where I can expand the svndiffs,
  ### I should also perform the digest calculation
  #import md5
  #m = md5.new()
  assert(length >= 4)
  offset = f.tell()
  header = f.read(4)
  if header != 'SVN\0':
    raise ValueError, 'No SVN diff present'
  length = length - (f.tell() - offset)
  print "starting length:", length
  print "offset:", offset

  total = 0
  window_id = 1
  try:
    while length > 0:
      bs = bitstream(f)
      offset = bs.f.tell()
      source_offset = get_varint(bs)
      source_length = get_varint(bs)
      target_length = get_varint(bs)
      instruction_length = get_varint(bs)
      new_data_length = get_varint(bs)
    
      ### Here we should expand the contents of the window
      ### and digest the data.  However, I'm only working
      ### with a single rev file, so I can't do this.
      #current = bs.f.tell()
      #bs.f.seek(offset)
      #m.update(bs.f.read(current - offset + instruction_length +
      #                   new_data_length))
      
      if dump_windows:
        print '  found window %d at %d:' % (window_id, offset)
        print '   source offset: %d' % source_offset
        print '   source length: %d' % source_length
        print '   target length: %d' % target_length
        print '   instr. length: %d' % instruction_length
        print '   data length:   %d' % new_data_length
      
      # parse the instructions
      total_data_bytes = 0
      total_target_bytes = 0
      tmp = instruction_length
      while tmp > 0:
        (instr_size, data_length, target_length_tmp) = \
          parse_svndiff_instruction(bs, dump_instructions)
        tmp = tmp - instr_size
        total_data_bytes = total_data_bytes + data_length
        total_target_bytes = total_target_bytes + target_length_tmp
      if dump_windows:
        print '   end of window: %d ' % (bs.f.tell()+new_data_length)

      if tmp < 0:
        print 'Decoded too many bytes'
        raise 'svndiff error'
  
      if total_data_bytes != new_data_length or \
         total_target_bytes + total_data_bytes != target_length:
        print 'total data bytes consumed (%d) != ' \
              'number of data bytes present (%d)' % \
                            (total_data_bytes, new_data_length)
        print 'total target bytes present (%d) != ' \
              'number of target bytes (%d)' % \
               (total_target_bytes + total_data_bytes, target_length)
        raise 'svndiff error'
  
      bs.f.seek(new_data_length, 1)
      size = bs.f.tell() - offset
      if size > length:
        print "size > length"
        print "size:", size
        print "length:", length
        raise "error"
      length = length - size
      total = total + size
      window_id = window_id + 1
  except:
    print "total:", total
    print "remaining:", length
    raise

  if length != 0:
    print "length remaining:", length
  if f.read(7) != 'ENDREP\n':
    print "Expected ENDREP at offset %d" % (f.tell()-7)

  ### Sould return the digest?
  return None


def get_dir_hash(f):
  l = f.readline()
  if l != 'PLAIN\n':
    raise ValueError, "Expected a PLAIN representation"
    
  hash = {}
  
  while True:
    field = f.readline()[:-1]
    if field == 'END':
      break
    assert(field[0] == 'K')
    length = int(field.split(' ')[1])
    field = f.readline()[:length]
    
    value = f.readline()[:-1]
    assert(value[0] == 'V')
    length = int(value.split(' ')[1])
    value = f.readline()[:length]
    
    (type, txn) = value.split(' ')
    hash[field] = [NodeType(type), NodeId(txn)]

  return hash


class Rep(object):
  def __init__(self, type, rev, offset, length, size, digest,
               content_type, current_rev, noderev):
    self.type = type
    self.rev = int(rev)
    self.offset = int(offset)
    self.length = int(length)
    self.size = int(size)
    self.digest = digest
    self.current_rev = current_rev

    self.content_type = content_type
    self.noderev = noderev

  def __repr__(self):
    if not self.content_type:
      content_type = 'UNKNOWN'
    else:
      if self.content_type not in ['PLAIN', 'DELTA', None]:
        content_type = 'INVALID'
      else:
        content_type = self.content_type
    return '%s: %s %d %d %d %d %s' % (self.type, content_type, self.rev,
                                      self.offset, self.length, self.size,
                                      self.digest)

  def verify(self, f, dump_instructions, dump_windows):
    if self.content_type not in ['PLAIN', 'DELTA', None]:
      raise FsfsInvalidHeaderError("Invalid rep header found at %d (%s)!" % \
                                     (self.offset, self.content_type),
                                   self.noderev, self, self.offset)
    f.seek(self.offset)
    header = f.read(5)
    if header != self.content_type:
      raise FsfsError, "Invalid rep header found at %d (%s)!" % (self.offset,
                                                                 header)
    if header == 'DELTA':
      while f.read(1) != '\n':
        pass
      digest = parse_svndiff(f, self.length, dump_instructions, dump_windows)
      if digest:
        assert(digest == self.digest)
    else:
      assert(f.read(1) == '\n')
      import md5
      m = md5.new()
      m.update(f.read(self.length))
      assert(self.digest == m.hexdigest())
      assert(f.read(7) == 'ENDREP\n')


class TextRep(Rep):
  def __init__(self, rev, offset, length, size, digest, 
               content_type, current_rev, noderev):
    super(TextRep,self).__init__('text', rev, offset, length, size,
                                 digest, content_type, current_rev, noderev)

    
class PropRep(Rep):
  def __init__(self, rev, offset, length, size, digest, 
               content_type, current_rev, noderev):
    super(PropRep,self).__init__('prop', rev, offset, length, size,
                                 digest, content_type, current_rev, noderev)

  
class NodeId(object):
  def __init__(self, nodeid):
    (self.txn_name, offset) = nodeid.split('/')
    self.offset = int(offset)
    self.rev = int(self.txn_name.split('.')[2][1:])

  def __repr__(self):
    return self.txn_name + '/%d' % self.offset

  def __eq__ (self, other):
    s = self.txn_name + '/%d' % self.offset
    if s == other:
      return True

    return False


class NodeType(object):
  def __init__(self, t):
    if (t != 'file') and (t != 'dir'):
      raise ValueError, 'Invalid Node type received: "%s"' % t
    self.type = t

  def __repr__(self):
    return self.type[:]


class NodeRev(object):
  def __init__(self, f, current_rev):
    self.pred = None
    self.text = None
    self.props = None
    self.cpath = None
    self.copyroot = None
    self.copyfrom = None
    self.dir = []

    self.node_offset = f.tell()

    while True:
      line = f.readline()
      if line == '':
        raise IOError, "Unexpected end of file"
      if line == '\n':
        break

      # break apart the line
      (field, value) = line.split(':', 1)

      # pull of the leading space and trailing new line
      value = value[1:-1]

      if field == 'id':
        self.id = NodeId(value)
      elif field == 'type':
        self.type = NodeType(value)
      elif field == 'pred':
        self.pred = NodeId(value)
      elif field == 'text':
        (rev, offset, length, size, digest) = value.split(' ')
        rev = int(rev)
        offset = int(offset)
        length = int(length)
        size = int(size)

        if rev != current_rev:
          content_type = None
        else:
          saved_offset = f.tell()
          f.seek(offset)
          content_type = f.read(5)
          f.seek(saved_offset)

        self.text = TextRep(rev, offset, length, size, digest, 
                            content_type, current_rev, self)
      elif field == 'props':
        (rev, offset, length, size, digest) = value.split(' ')
        rev = int(rev)
        offset = int(offset)
        length = int(length)
        size = int(size)

        if rev != current_rev:
          content_type = None
        else:
          saved_offset = f.tell()
          f.seek(offset)
          content_type = f.read(5)
          f.seek(saved_offset)

        self.props = PropRep(rev, offset, length, size, digest,
                             content_type, current_rev, self)
      elif field == 'cpath':
        self.cpath = value
      elif field == 'copyroot':
        self.copyroot = value
      elif field == 'copyfrom':
        self.copyfrom = value

    if self.type.type == 'dir':
      if self.text:
        if self.id.rev == self.text.rev:
          offset = f.tell()
          f.seek(self.text.offset)
          self.dir = get_dir_hash(f)
          f.seek(offset)
        else:
          # The directory entries are stored in another file.
          raise FsfsError, "dir entries are stored in rev %d" % self.text.rev

    if self.text:
      ### Verify the text rep here?
      pass

  def __repr__(self):
    str = 'NodeRev Id: %s\n type: %s\n' % (repr(self.id), repr(self.type))
    if self.pred:
      str = str + ' pred: %s\n' % repr(self.pred)
    if self.text:
      str = str + ' %s\n' % repr(self.text)
    if self.props:
      str = str + ' %s\n' % repr(self.props)
    if self.cpath:
      str = str + ' cpath: %s\n' % self.cpath
    if self.copyroot:
      str = str + ' copyroot: %s\n' % self.copyroot
    if self.copyfrom:
      str = str + ' copyfrom: %s\n' % self.copyfrom
    if self.dir:
      str = str + ' dir contents:\n'
      for k in self.dir:
        str = str + '  %s: %s\n' % (k, self.dir[k])
    return str[:-1]


def get_root_and_changed_paths(rev_file):
  offset = -2
  while True:
    rev_file.seek(offset, 2)
    c = rev_file.read(1)
    if c == '\n':
      offset = rev_file.tell()
      break
    offset = offset - 1
  
  (root_node, changed_paths) = map(int, rev_file.readline().split(' '))
  
  return (root_node, changed_paths)


class ChangedPaths(object):
  def __init__(self, f):
    self.changed_paths = {}

    while True:
      action = rev_file.readline()
      if action == '\n' or action == '':
        break

      path = action[:-1]
      (id, action, text_mod, prop_mod) = action[:-1].split(' ')[:4]
      path = path[len(' '.join([id, action, text_mod, prop_mod]))+1:]

      line = rev_file.readline()
      if line != '\n':
        (copyfrom_rev, copyfrom_path) = line.split(' ')
      else:
        copyfrom_rev = -1
        copyfrom_path = ''

      self.changed_paths[path] = (id, action, text_mod, prop_mod,
                                  copyfrom_rev, copyfrom_path)


  def __iter__(self):
    return self.changed_paths.iteritems()


class WalkStrategy(object):
  def __init__(self, filename, root_offset, current_rev):
    self.f = open(filename, 'rb')
    self.root_offset = root_offset
    self.f.seek(root_offset)
    self.current_rev = current_rev

  def _node_walker(self):
    raise NotImplementedError, "_node_walker is not implemented"

  def __iter__(self):
    self.f.seek(self.root_offset)
    return self._node_walker()


class ClassicStrategy(WalkStrategy):
  def _node_walker (self):
    noderev = NodeRev(self.f, self.current_rev)
    yield noderev

    if noderev.type.type == 'dir':
      for e in noderev.dir:
        if noderev.dir[e][1].rev == noderev.id.rev:
          self.f.seek(noderev.dir[e][1].offset)
          for x in self._node_walker():
            yield x


class RegexpStrategy(WalkStrategy):
  def _node_walker(self):
    import re
    node_id = re.compile(r'^id: [a-z0-9\./]*$')

    self.f.seek(0)
    offset = 0
    for line in self.f:
      match = node_id.search(line)
      if match:
        saved_offset = self.f.tell()
        self.f.seek(offset)
        noderev = NodeRev(self.f, self.current_rev)
        yield noderev
        self.f.seek(saved_offset)

      offset = offset + len(line)


def verify_noderev (rev_file, noderev, dump_instructions):
  if noderev.text and noderev.id.rev == noderev.text.rev:
    verify_rep(rev_file, noderev.text.offset, noderev.text.length,
               noderev.text.digest, dump_instructions)
  if noderev.props and noderev.id.rev == noderev.props.rev:
    verify_rep(rev_file, noderev.props.offset, noderev.props.length,
               noderev.props.digest, dump_instructions)


def check_options(options):
  count = 0
  for k,v in options.__dict__.items():
    if v and k != 'dump_changed':
      count = count + 1

  if count > 1:
    print >>sys.stderr, "Please use only one of -i, -n, and -r."
    sys.exit(1)


def dump_changed_paths(changed_paths):
  print "Changed Path Information:"
  for (path, 
       (id, action, text_mod, prop_mod, 
        copyfrom_rev, copyfrom_path)) in changed_paths:
    print " %s:" % path
    print "  action: %s" % action
    print "  text mod: %s" % text_mod
    print "  prop mod: %s" % text_mod
    if copyfrom_rev != -1:
      print "copyfrom path: %s" % copyfrom_path
      print "copyfrom rev: %s" % copyfrom_rev
    print


from optparse import OptionParser

parser = OptionParser("usage: %prog [-w | -i | -r | -n] REV-FILE")
parser.add_option("-c", "--changed-paths",
                  action="store_true", dest="dump_changed",
                  help="Dump changed path information", default=False)
parser.add_option("-i", "--instructions",
                  action="store_true", dest="dump_instructions",
                  help="Dump instructions (implies -w)", default=False)
parser.add_option("-w", "--windows",
                  action="store_true", dest="dump_windows",
                  help="Dump windows", default=False)
parser.add_option("-n", "--noderev-regexp",
                  action="store_true", dest="noderev_regexp",
                  help="Find all noderevs using a regexp", default=False)
parser.add_option("-r", "--representations",
                  action="store_true", dest="find_reps",
                  help="Find all representations using a regexp", default=False)
parser.add_option("-f", "--fix-read-length-line-error",
                  action="store_true", dest="fix_rlle",
                  help="Attempt to fix the mysterious read length line error",
                  default=False)
parser.add_option("-t", "--truncate",
                  action="store", type="string", dest="truncate",
                  help="Truncate the specified node rev.",
                  default=None)

(options, args) = parser.parse_args()

if len(args) != 1:
  print >>sys.stderr, "Please specify exactly one rev file."
  parser.print_help()
  sys.exit(1)

check_options(options)

filename = args[0]

if options.dump_instructions:
  options.dump_windows = True

if options.truncate or options.fix_rlle:
  rev_file = open(filename, 'r+b')
else:
  rev_file = open(filename, 'rb')
  
(root, changed) = get_root_and_changed_paths(rev_file)

if options.dump_changed:
  rev_file.seek(changed)
  changed_paths = ChangedPaths(rev_file)
  
  dump_changed_paths(changed_paths)
  sys.exit(0)

try:
  import re
  match = re.match('([0-9]+)', os.path.basename(filename))
  current_rev = int(match.group(1), 10)
except:
  raise CmdlineError, \
    "The file name must start with a decimal number that indicates the revision"

if options.noderev_regexp:
  strategy = RegexpStrategy(filename, root, current_rev)
else:
  strategy = ClassicStrategy(filename, root, current_rev)

# Make stderr the same as stdout.  This helps when trying to catch all of the
# output from a run.
sys.stderr = sys.stdout

def verify (noderev, rev_file, dump_instructions, dump_windows):
  print noderev
  if noderev.text:
    noderev.text.verify(rev_file,
                        dump_instructions,
                        dump_windows)
  print

def truncate (noderev, rev_file, dump_instructions, dump_windows):
  txn_id = noderev.id
  if txn_id == options.truncate:
    print "Truncating node %s (%s)" % (txn_id, noderev.cpath)

    # Grab the text rep
    text_rep = noderev.text

    # Fix the text rep contents
    offset = text_rep.offset
    rev_file.seek(offset, 0)
    rev_file.write("PLAIN\x0aENDREP\x0a")

    # Fix the node rev
    offset = noderev.node_offset
    rev_file.seek(offset, 0)
    while True:
      saved_offset = rev_file.tell()
      s = rev_file.readline()
      if s[:4] == 'text':
        rev_file.seek(saved_offset, 0)
        break

    line = rev_file.readline()
    rev_file.seek(saved_offset, 0)
    fields = line.split(' ')
    overall_length = len(line)

    fields[3] = '0' * len(fields[3])
    fields[4] = '0' * len(fields[4])
    fields[5] = 'd41d8cd98f00b204e9800998ecf8427e'
    new_text_rep = ' '.join(fields) + '\x0a'
    assert(len(new_text_rep) == overall_length)
    rev_file.write(new_text_rep)
    print "Done."

if not options.truncate:
  process = verify
else:
  process = truncate

try:
  for noderev in strategy:
    try:
      process(noderev, rev_file, options.dump_instructions, options.dump_windows)
    except:
      sys.stdout.flush()
      raise
except FsfsInvalidHeaderError, inv:
  # Looks like the read length line error problem... now what?
  print inv
  print "Found problem in NodeRev", inv.noderev.id, "(%d)" % inv.noderev.node_offset
  print inv.rep
  print "At offset", inv.offset

  if options.fix_rlle:
    # Let's give it a whirl!

    # First, we need to locate the real start of the text rep
    import re
    textrep_re = re.compile(r'^(DELTA( \d+ \d+ \d+)?|PLAIN)$')

    rev_file.seek(0)
    offset = 0
    original_offset = 0
    for line in rev_file:
      m = textrep_re.match(line)
      if m:
        if offset >= original_offset and offset < inv.offset:
          original_offset = offset
          header_len = len(line)
      offset = offset + len(line)

    print "Original text rep located at", original_offset

    # Okay, now we have the original offset of the text rep that was
    # in the process of being written out.  The header portion of the
    # text rep has a fsync() done after it, so the 4K blocks actually
    # start after the header.  We need to make sure to copy the header
    # and the next 4K, to be on the safe side.
    copy_len = 4096 + header_len

    rev_file.seek(original_offset)
    block = rev_file.read(copy_len)
    print "Copy %d bytes from offset %d" % (copy_len, original_offset)

    print "Write %d bytes at offset %d" % (copy_len, inv.offset)
    rev_file.seek(inv.offset)
    rev_file.write(block)
    rev_file.flush()

    print "Fixed? :-)  Re-run fsfsverify without the -f option"

